#ifndef _DOOLIB_BOUNDINGBOX2D_HEADER_
#define _DOOLIB_BOUNDINGBOX2D_HEADER_

#include "igeom.h"

#include <vector>
#include <algorithm>
#include <limits>
#include <ostream>

namespace diva
{

namespace imgproc
{

/*!\addtogroup ComputationalGeometryBaseClasses Base Classes
 * \ingroup ComputationalGeometry
 */
//@{
/*!BoundingBox2D represents the smallest axis-aligned rectangle surrounding a
 * certain graphical object. It provides several useful member functions, some of
 * which may be overridden in the derived classes (e.g. for implementing more exact
 * intersection tests).
 * TODO: change BoundingBox2D so it does not use closed intervals, but open ones - like in STL -
 * VERY useful for many algorithms - e.g. contours, polygons, etc as they all use this point of view...
 */
class BoundingBox2D
{
public:
    int minx, miny, maxx, maxy;

//-------------------------------------------------------------------------------
//                CONSTRUCTORS & DESTRUCTOR
//-------------------------------------------------------------------------------

    /*!Default constructor for class BoundingBox2D. A bounding box created using
     * the default constructor is initially invalid. (i.e. IsValid()==false)
     */
    BoundingBox2D() //:BasicObject()
    {
        InitDefaults();
    }

    BoundingBox2D(int _minx, int _miny, int _maxx, int _maxy) :
            /*BasicObject(),*/minx(_minx), miny(_miny), maxx(_maxx), maxy(_maxy)
    {
    }

    BoundingBox2D(const BoundingBox2D &b) :/*BasicObject(false),*/
            minx(b.minx), miny(b.miny), maxx(b.maxx), maxy(b.maxy)
    {
    }

    virtual ~BoundingBox2D()
    {
    }

//-------------------------------------------------------------------------------
//                STATIC FUNCTIONS
//-------------------------------------------------------------------------------

    /*!the distance on the OY axis between 2 BoundingBoxes
     * (if it's <0 => they overlap vertically)
     */
    static int disty(const BoundingBox2D* a, const BoundingBox2D* b)
    {
        int i = std::max(a->miny, b->miny) - std::min(a->maxy, b->maxy);
        return ((i <= 0) ? i - 1 : i);
    }

    /*!the distance on the OX axis between 2 BoundingBoxes
     * (if it's <0 => they overlap horizontally)
     */
    static int distx(const BoundingBox2D* a, const BoundingBox2D* b)
    {
        int i = std::max(a->minx, b->minx) - std::min(a->maxx, b->maxx);
        return ((i <= 0) ? i - 1 : i);
    }

    /*!Returns in res the BoundingBox2D which represents the intersection
     * of a and b. In case a and b don't intersect, res will be invalid.
     * \param[in] a
     * \param[in] b
     * \param[out] res
     * \pre a,b,res!=NULL and they point to valid BoundingBox2D objects
     * \sa IsValid()
     */
    static void GetIntersectionBox(const BoundingBox2D* a, const BoundingBox2D* b, BoundingBox2D* res)
    {
        res->minx = std::max(a->minx, b->minx);
        res->maxx = std::min(a->maxx, b->maxx);
        res->miny = std::max(a->miny, b->miny);
        res->maxy = std::min(a->maxy, b->maxy);
    }

    /*!Returns in res the BoundingBox2D which represents the intersection
     * of a and b. In case a and b don't intersect, res will be invalid.
     * \param[in] a
     * \param[in] b
     * \param[out] res
     * \pre a,b,res point to valid BoundingBox2D objects
     * \sa IsValid()
     */
    static void GetIntersectionBox(const BoundingBox2D& a, const BoundingBox2D& b, BoundingBox2D& res)
    {
        res.minx = std::max(a.minx, b.minx);
        res.maxx = std::min(a.maxx, b.maxx);
        res.miny = std::max(a.miny, b.miny);
        res.maxy = std::min(a.maxy, b.maxy);
    }

//-------------------------------------------------------------------------------
//                NORMAL MEMBER FUNCTIONS
//-------------------------------------------------------------------------------

    /*!Initializes all members of the BoundingBox with default values, so that
     * the BoundingBox is not Valid! (i.e. IsValid()==false)
     */
    void InitDefaults()
    {
        maxx = maxy = std::numeric_limits<int>::min();
        miny = minx = std::numeric_limits<int>::max();
    }

    /*!Returns true if the boundingbox represents a valid axis-aligned rectangle,
     * i.e. if the \c minx coordinate is <= \c maxx and the \c miny coordinate is <=
     * \c maxy.
     */
    bool IsValid() const
    {
        return ((minx <= maxx) && (miny <= maxy));
    }

    void operator=(const BoundingBox2D &b)
    {
//            this->BasicObject::operator=(*((const BasicObject*)(&b)));
        //not really worth testing if the assigned object is different than self...
        minx = b.minx;
        maxx = b.maxx;
        miny = b.miny;
        maxy = b.maxy;
    }

    /*!Performs a translation of the box coordinates using the given point (=vector).
     */
    void operator+=(const Point2D<int>& p)
    {
        minx += p.x;
        miny += p.y;
        maxx += p.x;
        maxy += p.y;
    }

    float GetXCenter() const
    {
        return (minx + maxx) * 0.5f;
    }

    float GetYCenter() const
    {
        return (miny + maxy) * 0.5f;
    }

    /*!Returns true if there is a vertical overlap between \b this and \b a, false
     * otherwise.
     * \pre a!=NULL, a->miny<=a->maxy, this->miny<=this->maxy
     */
    bool OverlapsVertically(const BoundingBox2D* a) const
    {
        return ((std::max(a->miny, this->miny) - std::min(a->maxy, this->maxy)) <= 0);
    }

    /*!Returns true if there is a horizontal overlap between \b this and \b a,
     * false otherwise.
     * \pre a!=NULL, a->minx<=a->maxx, this->minx<=this->maxx
     */
    bool OverlapsHorizontally(const BoundingBox2D* a) const
    {
        return ((std::max(a->minx, this->minx) - std::min(a->maxx, this->maxx)) <= 0);
    }

    bool Contains(int px, int py) const
    {
        if ((py >= miny) && (py <= maxy) && (px >= minx) && (px <= maxx)) return true;
        return false;
    }

    template<class T>
    bool Contains(const Point2D<T>* c) const
    {
        if ((c->y >= miny) && (c->y <= maxy) && (c->x >= minx) && (c->x <= maxx)) return true;
        return false;
    }

    bool Contains(const BoundingBox2D* c) const
    {
        if ((c->miny >= miny) && (c->maxy <= maxy) && (c->minx >= minx) && (c->maxx <= maxx)) return true;
        return false;
    }

    bool Intersects(const BoundingBox2D *c) const
    {
        if ((minx <= c->maxx) && (c->minx <= maxx) && (miny <= c->maxy) && (c->miny <= maxy)) return true;
        return false;
    }

    /*!Enlarges the BoundingBox so that the given point will be contained within its bounds.
     */
    void EnlargeBoxToContainPoint(int x, int y)
    {
        if (minx > x) minx = x;
        if (miny > y) miny = y;
        if (maxx < x) maxx = x;
        if (maxy < y) maxy = y;
    }

    /*!Enlarges the BoundingBox so that the given point will be contained within its bounds.
     * \pre p!=NULL
     */
    template<class T>
    void EnlargeBoxToContainPoint(const Point2D<T>* p)
    {
        if (minx > p->x) minx = p->x;
        if (miny > p->y) miny = p->y;
        if (maxx < p->x) maxx = p->x;
        if (maxy < p->y) maxy = p->y;
    }

    /*!The current bounding box will be extended (as little as possible) so as to
     * also contain the bounding box \c c .
     * \remarks Works correctly even if the current bounding box was not valid
     * before this call, in which case the result will be the bounding box of \c c .
     */
    void MergeBounds(const BoundingBox2D* c)
    {
        if (c == NULL || !c->IsValid()) return;

        if (IsValid())
        {
            minx = std::min(minx, c->minx);
            maxx = std::max(maxx, c->maxx);
            miny = std::min(miny, c->miny);
            maxy = std::max(maxy, c->maxy);
        }
        else
        {
            (*this) = (*c);
        }
    }

    /*!The current bounding box will be extended (as little as possible) so as to
     * also contain the bounding box \c c .
     * \remarks Works correctly even if the current bounding box was not valid
     * before this call, in which case the result will be the bounding box of \c c .
     */
    void MergeBounds(const BoundingBox2D& c)
    {
        if (!c.IsValid()) return;

        if (IsValid())
        {
            minx = std::min(minx, c.minx);
            maxx = std::max(maxx, c.maxx);
            miny = std::min(miny, c.miny);
            maxy = std::max(maxy, c.maxy);
        }
        else
        {
            (*this) = c;
        }
    }

    int Height() const
    {
        return (maxy - miny + 1);
    }

    int Width() const
    {
        return (maxx - minx + 1);
    }

    int Area() const
    {
        return ((maxx - minx + 1) * (maxy - miny + 1));
    }

    bool operator==(const BoundingBox2D& b) const
    {
        if ((minx == b.minx) && (miny == b.miny) && (maxx == b.maxx) && (maxy == b.maxy))
        {
            return true;
        }
        return false;
    }

    bool operator!=(const BoundingBox2D& b) const
    {
        return (!((*this) == b));
    }
};

std::ostream& operator<<(std::ostream& outstream, const BoundingBox2D& box);

/*!Base class for bounded (graphical) objects. It encapsulates a BoundingBox2D object
 * which represents an approximation of the graphical object's border and a few
 * access member functions.
 */
class BoundedObject
{
protected:
    BoundingBox2D bbox;

public:
    BoundedObject()
    {
    }

    BoundedObject(const BoundedObject& bobj) :
            bbox(bobj.bbox)
    {
    }

    const BoundingBox2D* GetBoundingBox() const
    {
        return &bbox;
    }

    BoundingBox2D* GetBoundingBox()
    {
        return &bbox;
    }

    virtual bool operator==(const BoundedObject& bobj) const
    {
        if (bbox == bobj.bbox)
        {
            return true;
        }
        return false;
    }

    virtual bool operator!=(const BoundedObject& bobj) const
    {
        return (!((*this) == bobj));
    }

    virtual void operator=(const BoundedObject& bobj)
    {
        if (this == &bobj)
        {
            return;
        }
        bbox = bobj.bbox;
    }

    virtual ~BoundedObject()
    {
    }
};

std::ostream& operator<<(std::ostream& outstream, const BoundedObject& bobj);

//@}

//---------------------------------------------------------------------------------------
//                    implementations of INLINE/ TEMPLATE functions
//---------------------------------------------------------------------------------------

inline std::ostream& operator<<(std::ostream& outstream, const BoundingBox2D& box)
{
    outstream << "BBox: (" << box.minx << ", " << box.miny << ") -> (" << box.maxx << ", " << box.maxy << ")";
    return outstream;
}

inline std::ostream& operator<<(std::ostream& outstream, const BoundedObject& bobj)
{
    outstream << "BObj: (" << bobj.GetBoundingBox()->minx << ", " << bobj.GetBoundingBox()->miny << ") -> (" << bobj.GetBoundingBox()->maxx << ", "
            << bobj.GetBoundingBox()->maxy << ")";
    return outstream;
}

} //namespace image

}

#endif //ndef _DOOLIB_BOUNDINGBOX2D_HEADER_
