#ifndef _DOO_COLOR_H_
#define _DOO_COLOR_H_

#include "ipixel_types.h"

#include <cassert>
#include <limits>
#include <cstddef>
#include <string>
#include <vector>

namespace diva
{

namespace imgproc
{

/*!\addtogroup ColorSpaceConversions Color Space Conversions
 * \ingroup ImageManipulation
 */

/*!\addtogroup Colors Color Management
 */

/*!\addtogroup ColorBasic Basic Color Manipulation, Color Spaces and Traits
 * \ingroup Colors
 * \brief Include header file(s): <doolib/icolor.h>
 */
//@{
/*!32-bitcolor value consisting of 8-bit "unsigned chars": alpha, red, green, blue.
 * \note alpha = 255 <=> fully opaque
 */
typedef uint32_t ARGB;

ARGB MakeARGB(const int a, const int r, const int g, const int b);

/*!Enumeration of all supported colorspaces.
 * It defines the colormodel and a default colorspace. If the colormodel is used with
 * a different colorpace, than an ICC-profile have to be provided in the
 * element icc of the image.
 * \warning LAB is CIE L*a*b*, not HunterLAB
 */
enum COLORSPACE
{
    GRAYSCALE,          //!< luma (Y) using ITU-R BT.601 conversion from sRGB
    RGB,                //!< sRGB
    YCBCR,              //!< (JPEG conversion) an offset of 128 is used to fit Cb and Cr in the range 0..255
    LAB,                //!< Whitepoint D50 2°, all values range from 0 to 255 for PIX (factor 2.55 for L, offset of 128 for a and b)
    CMYK,               //!< derived from sRGB
    XYZ,                //!< Whitepoint D50 2°
    COLORSPACE_COUNT
};

/*!Class containing information about all supported colorspaces and conversions.
 */
class ColorSpaceTraits
{
public:

    static bool IsSpecialized(const COLORSPACE cspace);
    static bool IsRGBConnected(const COLORSPACE cspace);
    static std::size_t GetDimension(const COLORSPACE cspace);
    static std::string GetStringID(const COLORSPACE cspace);
};

//@}

//---------------------------------------------------------------------------------------
//                    implementations of INLINEs/ TEMPLATEs
//---------------------------------------------------------------------------------------

inline ARGB MakeARGB(const int a, const int r, const int g, const int b)
{
    assert((std::numeric_limits<PIX>::min() <= a) && (a <= std::numeric_limits<PIX>::max()));
    assert((std::numeric_limits<PIX>::min() <= r) && (r <= std::numeric_limits<PIX>::max()));
    assert((std::numeric_limits<PIX>::min() <= g) && (g <= std::numeric_limits<PIX>::max()));
    assert((std::numeric_limits<PIX>::min() <= b) && (b <= std::numeric_limits<PIX>::max()));

    return (((ARGB(PIX(a))) << 24) | ((ARGB(PIX(r))) << 16) | ((ARGB(PIX(g))) << 8) | (ARGB(PIX(b))));
}

}

}

#endif //ndef _IMAGE_COLOR_H_
