#ifndef _DOO_CONCOMP_HEADER
#define _DOO_CONCOMP_HEADER

#include "image.h"
#include "boundingbox2d.h"

#include <cstddef>

namespace diva
{

namespace imgproc
{

/*!\defgroup ConComps Connected Components
 * \brief Include header file(s): <doolib/iconcomp.h>
 */
//@{
/*!Connected component.
 */
class ConComp: public BoundedObject
{
public:
    float centerx, centery; //center of mass (NOT center of bounding box)
    int numpixel;
    int label;

    union
    {
        uint32_t grayvalue;
        ARGB argb;
    };

    ConComp()
    {
        centerx = centery = 0.0;
        numpixel = 0;
        grayvalue = 0;
        label = -1;
    }

    ConComp(const ConComp& c) : BoundedObject()
    {
        *this = c;
    }

    ~ConComp()
    {
    }

    void operator=(const ConComp& c)
    {
        if (this == &c)
        {
            return;
        }
        this->BoundedObject::operator=(c);
        centerx = c.centerx;
        centery = c.centery;
        numpixel = c.numpixel;
        grayvalue = c.grayvalue;
        label = c.label;
    }

    inline int Normalize()
    {
        if (numpixel == 0) return (-1);
        centerx /= numpixel;
        centery /= numpixel;
        return 0;
    }
};

unsigned int CreateConCompArray(const IMAGE *img, const IIMAGE *labelimage, ConComp** complist);
unsigned int LabelImage(const IMAGE *img, IIMAGE **plabelimg);
unsigned int LabelImageN8(const IMAGE *img, IIMAGE **plabelimg);

/*!Contains the identifiers of all the implemented neighborhood types for
 * region labeling.
 */
enum NeighborhoodType
{
    NEIGHBORHOOD_UNKNOWN, NEIGHBORHOOD_4_CONNECTED, NEIGHBORHOOD_8_CONNECTED, NEIGHBORHOOD_TYPES_COUNT //NOTE: NEVER use NEIGHBORHOOD_TYPES_COUNT as a region type! - it exists only
                                                                                                       //for being able to determine the number of possible neighborhood types
};

/*!Contains the labeling and the connected components for a certain image.
 */
class ConnectedComponents
{
public:
    union
    {
        IIMAGE *labelimage;
        IIMAGE *labimg;
    };

    union
    {
        unsigned int numcomps;  // deprecated
        unsigned int numcomp;
    };

    ConComp *components;

    union
    {
        NeighborhoodType usedNeighborhood;  // deprecated
        NeighborhoodType neighborhood;   //!< can only be set when a labeling is actually performed/ initialized
    };

    ConnectedComponents()
    {
        initDefaults();
    }

    ConnectedComponents(const IMAGE& img, NeighborhoodType useNeighborhood = NEIGHBORHOOD_4_CONNECTED)
    {
        initDefaults();
        Relabel(img, useNeighborhood);
    }

    ConnectedComponents(IIMAGE* labimg, ConComp *comps, unsigned int ncomps, NeighborhoodType used_neighborhood = NEIGHBORHOOD_4_CONNECTED)
    {
        if (labimg == NULL || comps == NULL)
        {
            initDefaults();
        }
        else
        {
            labelimage = labimg;
            usedNeighborhood = used_neighborhood;

            numcomps = ncomps;
            components = comps;
        }
    }

    ~ConnectedComponents()
    {
        Release();
    }

    void Release()
    {
        if (labelimage != NULL) delete labelimage;
        labelimage = NULL;
        delete[] components;
        components = NULL;
        initDefaults();
    }

    void initDefaults()
    {
        numcomps = 0;
        components = NULL;
        labelimage = NULL;
        usedNeighborhood = NEIGHBORHOOD_UNKNOWN;
    }

    unsigned int Relabel(const IMAGE& img, NeighborhoodType use_neighborhood = NEIGHBORHOOD_4_CONNECTED)
    {
        if (labelimage == NULL) labelimage = new IIMAGE(img.xdim, img.ydim, GRAYSCALE);

        if (labelimage->xdim != img.xdim || labelimage->ydim != img.ydim)
        {
            delete labelimage;
            labelimage = new IIMAGE(img.xdim, img.ydim, GRAYSCALE);
        }

        delete[] components;

        if (use_neighborhood == NEIGHBORHOOD_8_CONNECTED)
        {
            /*numcomps = */LabelImageN8(&img, &labelimage);
            usedNeighborhood = NEIGHBORHOOD_8_CONNECTED;
        }
        else    //assume 4-connectedness is wanted...
        {
            /*numcomps = */LabelImage(&img, &labelimage);
            usedNeighborhood = NEIGHBORHOOD_4_CONNECTED;
        }
        numcomps = CreateConCompArray(&img, labelimage, &components);

        return numcomps;
    }

    //ACCESSORS
    NeighborhoodType getUsedNeighborhood() const
    {
        return usedNeighborhood;
    }

    const ConComp* getComponents() const
    {
        return components;
    }

    ConComp* getComponents()
    {
        return components;
    }

    unsigned int getComponentsSize() const
    {
        return numcomps;
    }

    int getXDim() const
    {
        return ((labelimage == NULL) ? 0 : labelimage->xdim);
    }

    int getYDim() const
    {
        return ((labelimage == NULL) ? 0 : labelimage->ydim);
    }

    const IIMAGE* getLabelImage() const
    {
        return labelimage;
    }

    IIMAGE* getLabelImage()
    {
        return labelimage;
    }
};

//@}

}//namespace image

}

#endif    //CONCOMP_HEADER
