#ifndef _DOO_IDELAUNAY_H_
#define _DOO_IDELAUNAY_H_

#include "igeom.h"
#include "image.h"
#include "simpleallocator.h"

#include <vector>

namespace diva
{

namespace imgproc
{

/*!\defgroup Delaunay_Triangulations Delaunay Triangulations
 * \brief Include header file(s): <doolib/idelaunay.h>
 */

//!\addtogroup Delaunay_Triangulations
//@{
typedef enum
{
    delaunay_right, delaunay_left
} delaunay_side;

class edge_fwe;

/*!Point class for use within a full winged edge data structure.
 */
class point_fwe
{
public:
    double x;
    double y;
    edge_fwe *entry_pt;

    point_fwe(double xc = 0.0, double yc = 0.0, edge_fwe *entrypt = NULL) :
            x(xc), y(yc), entry_pt(entrypt)
    {
    }
};

/*!Vector class for use as utility in the delaunay triangulation algorithm.
 */
class vector_fwe
{
public:
    double x;
    double y;

    vector_fwe(double xc = 0.0, double yc = 0.0) :
            x(xc), y(yc)
    {
    }

    vector_fwe(const point_fwe& p1, const point_fwe& p2) :
            x(p2.x - p1.x), y(p2.y - p1.y)
    {
    }
};

/*!Edge class for use within a full winged edge data structure.
 */
class edge_fwe
{
public:
    point_fwe *org;
    point_fwe *dest;
    edge_fwe *onext;
    edge_fwe *oprev;
    edge_fwe *dnext;
    edge_fwe *dprev;

    edge_fwe()
    {
        // edge points to itself
        onext = this;
        oprev = this;
        dnext = this;
        dprev = this;

        org = NULL;
        dest = NULL;
    }
};

/*!Class for storing a polygon mesh using the full winged edge data structure.
 * \remarks Currently it contains no information about faces.
 */
class PolygonMeshFullWingedEdge
{
public:
#if defined(WIN32)
#pragma warning(disable: 4251)
#endif
    std::vector<edge_fwe> e_array;
    std::vector<point_fwe> p_array;
#if defined(WIN32)
#pragma warning(default: 4251)
#endif
};

PolygonMeshFullWingedEdge* DelaunayTriangulate2D(const std::vector<Point2D<double> >& sites);
//@}

//some other utility functions, not interesting from the outside
void ComputeLowerTangent(edge_fwe *r_cw_l, point_fwe *s, edge_fwe *l_ccw_r, point_fwe *u, edge_fwe *&l_lower, point_fwe *&org_l_lower, edge_fwe *&r_lower,
        point_fwe *&org_r_lower);
void MergeTriangulations(SimpleAllocator<edge_fwe>& e_alloc, edge_fwe *r_cw_l, point_fwe *s, edge_fwe *l_ccw_r, point_fwe *u, edge_fwe **l_tangent);
void SplitPointSet(SimpleAllocator<edge_fwe>& e_alloc, point_fwe *p_sorted[], int l, int r, edge_fwe *&l_ccw, edge_fwe *&r_cw);

//----------------------------------------------------------------------------
//                         IMPLEMENTATIONS of INLINE functions
//----------------------------------------------------------------------------

inline double CrossProduct(const vector_fwe &a, const vector_fwe &b)
{
    return (a.x * b.y - b.x * a.y);
}

inline double DotProduct(const vector_fwe &a, const vector_fwe &b)
{
    return (a.x * b.x + a.y * b.y);
}

inline double CrossProduct(const point_fwe& p1, const point_fwe& p2, const point_fwe& p3)
{
    return ((p2.x - p1.x) * (p3.y - p1.y) - (p2.y - p1.y) * (p3.x - p1.x));
}

inline point_fwe* Other_point(edge_fwe* e, point_fwe* p)
{
    return ((e)->org == p ? (e)->dest : (e)->org);
}

inline point_fwe* Other_point(edge_fwe const * const e, point_fwe const * const p)
{
    return ((e)->org == p ? (e)->dest : (e)->org);
}

inline edge_fwe* Next(edge_fwe* e, point_fwe* p)
{
    return ((e)->org == p ? (e)->onext : (e)->dnext);
}

inline edge_fwe* Prev(edge_fwe* e, point_fwe* p)
{
    return ((e)->org == p ? (e)->oprev : (e)->dprev);
}

/*!Initializes a new edge.
 */
inline edge_fwe *CreateEdgeFWE(SimpleAllocator<edge_fwe>& e_alloc, point_fwe *u, point_fwe *v)
{
    edge_fwe *e = e_alloc.NewObject();

    e->onext = e->oprev = e->dnext = e->dprev = e;
    e->org = u;
    e->dest = v;
    if (u->entry_pt == NULL)
    {
        u->entry_pt = e;
    }
    if (v->entry_pt == NULL)
    {
        v->entry_pt = e;
    }

    return e;
}

/*!Removes an edge from the full winged edge data structure.
 */
inline void DiscardEdgeFWE(SimpleAllocator<edge_fwe>& e_alloc, edge_fwe *e)
{
    point_fwe *u = e->org;
    point_fwe *v = e->dest;

    if (u->entry_pt == e)
    {
        u->entry_pt = e->onext;
    }
    if (v->entry_pt == e)
    {
        v->entry_pt = e->dnext;
    }

    if (e->onext->org == u)
    {
        e->onext->oprev = e->oprev;
    }
    else
    {
        e->onext->dprev = e->oprev;
    }

    if (e->oprev->org == u)
    {
        e->oprev->onext = e->onext;
    }
    else
    {
        e->oprev->dnext = e->onext;
    }

    if (e->dnext->org == v)
    {
        e->dnext->oprev = e->dprev;
    }
    else
    {
        e->dnext->dprev = e->dprev;
    }

    if (e->dprev->org == v)
    {
        e->dprev->onext = e->dnext;
    }
    else
    {
        e->dprev->dnext = e->dnext;
    }

    e_alloc.DeleteObject(e);
}

/*!Inserts an edge into a ring of edges.
 */
inline void SpliceEdgeChains(edge_fwe *a, edge_fwe *b, point_fwe *v)
{
    edge_fwe *next;

    if (a->org == v)
    {
        next = a->onext;
        a->onext = b;
    }
    else
    {
        next = a->dnext;
        a->dnext = b;
    }

    if (next->org == v)
    {
        next->oprev = b;
    }
    else
    {
        next->dprev = b;
    }

    if (b->org == v)
    {
        b->onext = next;
        b->oprev = a;
    }
    else
    {
        b->dnext = next;
        b->dprev = a;
    }
}

/*!Creates a new edge and adds it to two rings of edges.
 */
inline edge_fwe *JoinEdgeChains(SimpleAllocator<edge_fwe>& e_alloc, edge_fwe *a, point_fwe *u, edge_fwe *b, point_fwe *v, delaunay_side s)
{
    edge_fwe *e;

    e = CreateEdgeFWE(e_alloc, u, v);

    if (s == delaunay_left)
    {
        if (a->org == u)
        {
            SpliceEdgeChains(a->oprev, e, u);
        }
        else
        {
            SpliceEdgeChains(a->dprev, e, u);
        }
        SpliceEdgeChains(b, e, v);
    }
    else
    {
        SpliceEdgeChains(a, e, u);
        if (b->org == v)
        {
            SpliceEdgeChains(b->oprev, e, v);
        }
        else
        {
            SpliceEdgeChains(b->dprev, e, v);
        }
    }

    return e;
}

}

}

#endif //_DOO_IDELAUNAY_H_
