#ifndef _DOO_GEOMETRICAL_FUNCTIONS_HEADER
#define _DOO_GEOMETRICAL_FUNCTIONS_HEADER

//standard
#include <cassert>
#include <cstddef>
#include <utility>
#include <algorithm>

namespace diva
{

namespace imgproc
{

/*!\addtogroup ComputationalGeometry Computational Geometry-related
 */

/*!\addtogroup ComputationalGeometryBaseClasses Base Classes
 * \ingroup ComputationalGeometry
 */
//@{
/*!Class representing a 2-D point with coordinates of type \c SignedArithmeticType.
 * \remarks One MUST use a signed basic C++ type for SignedArithmeticType (int or double are
 * recommended) !.
 * \warning When using integer coordinates, beware of possible range overflows
 * in the geometrical functions !
 * \note Implements Point2DConcept.
 * \pre SignedArithmeticType implements the concept SignedArithmeticType
 * \sa Point2DConcept, SignedArithmeticType
 */
template<class SignedArithmeticType>
class Point2D
{
public:
    typedef SignedArithmeticType value_type;

    value_type x, y;

    explicit Point2D(SignedArithmeticType xi = 0, SignedArithmeticType yi = 0) :
            x(xi), y(yi)
    {
    }

    Point2D(const Point2D<SignedArithmeticType>& p) :
            x(p.x), y(p.y)
    {
    }

    template<class SignedArithmeticType2>
    explicit Point2D(const Point2D<SignedArithmeticType2>& p) :
            x(static_cast<SignedArithmeticType>(p.x)), y(static_cast<SignedArithmeticType>(p.y))
    {
    }

    std::size_t size() const
    {
        return std::size_t(2);
    }

    const value_type& operator[](std::size_t index) const
    {
        assert(index < 2);

        return this->*mem_array[index];
    }

    value_type& operator[](std::size_t index)
    {
        assert(index < 2);

        return this->*mem_array[index];
    }

    bool operator==(const Point2D<SignedArithmeticType>& p) const
    {
        return ((x == p.x) && (y == p.y));
    }

    bool operator!=(const Point2D<SignedArithmeticType>& p) const
    {
        return (!((*this) == p));
    }

    bool operator<(const Point2D<SignedArithmeticType>& p) const
    {
        if (x != p.x) return (x < p.x);
        return (y < p.y);
    }

    const Point2D<SignedArithmeticType>& operator=(const Point2D<SignedArithmeticType>& p)
    {
        x = p.x;
        y = p.y;

        return (*this);
    }

    /*!Efficiently swaps the current point with the given one.
     */
    void swap(Point2D<SignedArithmeticType>& p)
    {
        std::swap(x, p.x);
        std::swap(y, p.y);
    }

protected:
    //this static array of pointers to member variables makes operator[] "safe"
    //and doesn't seem to exhibit any performance penalty
    //note: this part/idea from the code comes from boost::gil
    static SignedArithmeticType Point2D<SignedArithmeticType>::* const mem_array[2];
};

//@}

template<typename SignedArithmeticType>
SignedArithmeticType Point2D<SignedArithmeticType>::* const Point2D<SignedArithmeticType>::mem_array[2] =
{ &Point2D<SignedArithmeticType>::x, &Point2D<SignedArithmeticType>::y };

/*!\addtogroup ComputationalGeometry Computational Geometry-related
 */
//@{
template<class SignedArithmeticType>
void swap(Point2D<SignedArithmeticType>& p1, Point2D<SignedArithmeticType>& p2);

//explicit exports of several template instantiations
template class Point2D<double> ;
template class Point2D<float> ;

//@}

/*!\addtogroup ComputationalGeometryBaseClasses Base Classes
 * \ingroup ComputationalGeometry
 */
//@{
template class Point2D<int> ;
typedef Point2D<int> IPOINT;

//@}

//-----------------------------------------------------------------------------------------
//				implementation of template & (possibly) inline functions
//-----------------------------------------------------------------------------------------

/*!Efficiently swaps 2 2-dimensional points.
 */
template<class SignedArithmeticType>
void swap(Point2D<SignedArithmeticType>& p1, Point2D<SignedArithmeticType>& p2)
{
    p1.swap(p2);
}

}

}
#endif
