#ifndef _DOO_NUMERICS_H_
#define _DOO_NUMERICS_H_

#include "numeric_constants.h"

namespace diva
{

namespace imgproc
{

/*!\addtogroup NumericFunctions Numeric Classes, Functions, Constants
 * \brief Include header file(s): <doolib/numerics.h>.
 */
//@{
template<class ArithmeticType>
ArithmeticType SQR(const ArithmeticType val);

template<class ValueType>
ValueType abs(ValueType val);

//@}

/*!\addtogroup UnitConversions Measurement Unit Conversions
 * \ingroup NumericFunctions
 */
//@{
template<class ArithmeticType>
ArithmeticType DegreesToRadians(const ArithmeticType val);

template<class ArithmeticType>
ArithmeticType RadiansToDegrees(const ArithmeticType val);
//@}

//---------------------------------------------------------------------------------
//             implementations of various inline/template functions
//---------------------------------------------------------------------------------

template<class ArithmeticType>
inline ArithmeticType SQR(const ArithmeticType val)
{
    return val * val;
}

/*!Converts the given value from degrees to radians.
 * \warning Make \b SURE the type of the input value is floating-point, otherwise the result
 * will be truncated to have the same type as \c val !
 */
template<class ArithmeticType>
inline ArithmeticType DegreesToRadians(const ArithmeticType val)
{
    return ArithmeticType(PI * val * _1_DIV_180);
}

/*!Converts the given value from radians to degrees.
 * \warning Make \b SURE the type of the input value is floating-point, otherwise the result
 * will be truncated to have the same type as \c val !
 */
template<class ArithmeticType>
inline ArithmeticType RadiansToDegrees(const ArithmeticType val)
{
    return ArithmeticType(val * 180.0 * _1_DIV_PI);
}

/*!Returns the absolute value of \c val.
 * \remarks Defined here as a template function because in standard C/C++ there
 * exists no uniform handling for it (the abs functions are split in \<cmath\> and
 * \<cstdlib\>, are NOT templates + there is no support (before C99, which is not
 * yet included in the C++ standard) for 64 bit integers.
 * \pre \c ValueType must be an ArithmeticType
 * \sa ArithmeticType
 */
template<class ValueType>
inline ValueType abs(ValueType val)
{
    return ((val < static_cast<ValueType>(0)) ? (-val) : (val));
}

}   //namespace

}

#endif

