#ifndef _DOO_SIMPLEALLOCATOR_H_
#define _DOO_SIMPLEALLOCATOR_H_

#include "ierror_handling.h"

#include <vector>

namespace diva
{

namespace imgproc
{

/*!\addtogroup Auxiliary_general Auxiliary classes and functions (general)
 */
//@{
/*!A convenience template class, which allows efficient dynamic allocation/deallocation
 * for objects of a certain type. It is very useful when many single objects
 * have to be dynamically created/deleted, and the maximum number of objects existent
 * at any time is known in advance.
 * \warning In order for the class to function properly, one must provide an EXACT
 * estimation (or an overestimation) for the maximum possible total no. of objects to be allocated.
 * \remarks REQUIREMENTS: The stored class should provide a default constructor,
 * as well as be a valid C++ STL container element (i.e. has copy constructor, destructor&
 * is assignable).
 */
template<class T>
class SimpleAllocator
{
public:
    std::vector<T> *objstore;
    std::vector<T*> *freeslots;

    SimpleAllocator(unsigned nmax)
    {
        objstore = new std::vector<T>(nmax);
        freeslots = new std::vector<T*>();
        freeslots->reserve(nmax);
        for (unsigned i = 0; i < nmax; ++i)
        {
            freeslots->push_back(&((*objstore)[i]));
        }
    }

    T* NewObject()
    {
        if (freeslots->empty())
        {
            ErrorImg("Out of memory for objects in SimpleAllocator => estimation was not correct !\n");
        }
        T* obj = freeslots->back();
        freeslots->pop_back();
        return obj;
    }

    void DeleteObject(T* e)
    {
        freeslots->push_back(e);
    }

    ~SimpleAllocator()
    {
        delete objstore;
        delete freeslots;
    }

protected:
    SimpleAllocator() :
            objstore(0), freeslots(0)
    {
    }
};

//@}

}//namespace

}

#endif //_DOO_SIMPLEALLOCATOR_H_
